package img.device;

import java.awt.*;
import java.io.*;
import java.nio.*;
import java.util.*;

import javax.comm.*;

import img.connection.*;
import img.device.event.*;
import img.io.*;

import img.lang.*;

public class Device {

   public static void main(String[] args) throws Exception {
      Device device = new Device();

      SerialParams params = new SerialParams();

      device.setConnection(new SerialConnection("COM1", params));

      device.connect();
      device.sendPing();
      device.sendDumpFrame();
      device.sendGetVersion();
      device.sendSetRegisters(new HashMap());
      device.sendReset();
      device.sendEnableTracking();
      device.sendDisableTracking();
      device.sendSetColorMap(null, null, null);

      try {
         System.in.read();
      } catch(IOException ioe) {
         ioe.printStackTrace();
      }
      device.disconnect();

   }

   private static final String ACK = "ACK";
   private static final String NCK = "NCK";
   private static final String VERSION = "AVR";

   private java.util.List dataListeners;
   private Connection con;

   private InputStream in;
   private OutputStream out;

   private int[][] colorMap;
   private Color[] mapColors;

   public Device() {
      dataListeners = new ArrayList(3);
      loadMap();
   }

   public Color[] getMapColors() {
      return mapColors;
   }

   public int[][] getColorMap() {
      return colorMap;
   }

   public void setColorMap(int[][] colorMap) {
      this.colorMap = colorMap;
      setMapColors();
      saveMap();
   }

   public void saveMap() {
      try {
         ByteArrayOutputStream bOut = new ByteArrayOutputStream();
         DataOutputStream dOut = new DataOutputStream(bOut);

         dOut.writeInt(colorMap.length);
         for(int i = 0; i < colorMap.length; i++) {
            dOut.writeInt(colorMap[i].length);
            for(int j = 0; j < colorMap[i].length; j++) {
               dOut.writeInt(colorMap[i][j]);
            }
         }

         dOut.close();

         ImageSystem.PREFS.putByteArray("img.color.map", bOut.toByteArray());

      } catch(Exception e) {
         e.printStackTrace();
      }
   }

   public void loadMap() {
      try {
         byte[] data = ImageSystem.PREFS.getByteArray("img.color.map", null);
         if(data != null) {
            ByteArrayInputStream bIn = new ByteArrayInputStream(data);
            DataInputStream dIn = new DataInputStream(bIn);

            int width = dIn.readInt();
            colorMap = new int[width][];
            for(int i = 0; i < colorMap.length; i++) {
               colorMap[i] = new int[dIn.readInt()];
               for(int j = 0; j < colorMap[i].length; j++) {
                  colorMap[i][j] = dIn.readInt();
               }
            }

            dIn.close();
         } else {
            colorMap = new int[3][ImageSystem.NUM_INTENSITIES];
         }

         setMapColors();

      } catch(Exception e) {
         e.printStackTrace();
      }

   }

   private void setMapColors() {
      mapColors = new Color[8];

      for(int col = 0; col < mapColors.length; col++) {
         int value = 0;

         int red = 0;
         int green = 0;
         int blue = 0;

         for(int i = 0; i < ImageSystem.NUM_INTENSITIES; i++) {

            if(red == 0 && (colorMap[0][i] & (0x01 << (7 - col))) != 0) {
               red = i << 4;
            }

            if(green == 0 && (colorMap[1][i] & (0x01 << (7 - col))) != 0) {
               green = i << 4;
            }

            if(blue == 0 && (colorMap[2][i] & (0x01 << (7 - col))) != 0) {
               blue = i << 4;
            }

         }

         value = (red << 16) | (green << 8) | blue;

         mapColors[col] = new Color(value);
      }

   }

   public Connection getConnection() {
      return con;
   }

   public void setConnection(Connection con) {
      this.con = con;
   }

   public boolean isConnected() {
      return con != null && con.isConnected();
   }

   public void connect() throws Exception {

      if(con != null && !con.isConnected()) {
         con.connect();

         if(con instanceof SerialConnection) {
            SerialPort serialPort = (SerialPort)con.getConnectionObject();
            serialPort.notifyOnDataAvailable(true);
            try {
               serialPort.addEventListener(new SerialEventHandler(this));
            } catch(TooManyListenersException tmle) {
               ImageSystem.LOG.severe(tmle.getMessage());
            }
         }

         in = new AVRInputStream(con.getInputStream());
         out = con.getOutputStream();

         ImageSystem.LOG.config("Device connected to " + con.toString());

      }
   }

   public void disconnect() {

      if(con != null && con.isConnected()) {
         try {
            con.disconnect();
            con = null;
            ImageSystem.LOG.config("Device Disconnected");
         } catch(IOException ioe) {
            ImageSystem.LOG.severe(ioe.getMessage());
         }
      }
   }

   public InputStream getInputStream() {
      return in;
   }

   public void addDataListener(DataListener dl) {
      if(dataListeners.add(dl)) {
         ImageSystem.LOG.finest("Added: " + dl);
      }
   }

   public void removeDataListener(DataListener dl) {
      if(dataListeners.remove(dl)) {
         ImageSystem.LOG.finest("Removed: " + dl);
      }
   }

   protected void handleString(String data) {
      if(data.equals(ACK)) {
         fireACKReceived();
      } else if(data.equals(NCK)) {
         fireNCKReceived();
      } else if(data.startsWith(VERSION)) {
         fireVERSIONReceived(data);
      } else {
         StringBuffer builder = new StringBuffer("UNKNOWN PACKET: (");
         ByteBuffer bytes = ByteBuffer.wrap(data.getBytes());
         builder.append(bytes.remaining()).append(") ");
         while(bytes.hasRemaining()) {
            builder.append(Integer.toHexString(bytes.get() & 0xFF)).append(' ');
         }
         ImageSystem.LOG.warning(builder.toString());
      }
   }

   protected void handleData(ByteBuffer data) {
      if(data.hasRemaining()) {
         byte dataType = data.get();
         if(dataType == 0x0B) {
            fireFrameDataReceived(data);
         } else if(dataType == 0x0A) {
            fireTrackingDataReceived(data);
         } else {
            handleString(new String(data.array(), 0, data.limit()));
         }
      }
   }

   protected void fireACKReceived() {
      ImageSystem.LOG.info("Received: ACK");
      DataListener[] listeners = (DataListener[])dataListeners.toArray(new DataListener[dataListeners.size()]);
      for(int i = 0; i < listeners.length; i++) {
         ((DataListener)listeners[i]).ack();
      }
   }

   protected void fireNCKReceived() {
      ImageSystem.LOG.info("Received: NCK");
      DataListener[] listeners = (DataListener[])dataListeners.toArray(new DataListener[dataListeners.size()]);
      for(int i = 0; i < listeners.length; i++) {
         ((DataListener)listeners[i]).nck();
      }
   }

   protected void fireVERSIONReceived(String version) {
      ImageSystem.LOG.info("Received: " + version);
      DataListener[] listeners = (DataListener[])dataListeners.toArray(new DataListener[dataListeners.size()]);
      for(int i = 0; i < listeners.length; i++) {
         ((DataListener)listeners[i]).version(version);
      }
   }

   protected void fireFrameDataReceived(ByteBuffer data) {
      ImageSystem.LOG.info("Received: Frame Dump (" + (data.get(data.position()) & 0xFF) + ")");
      DataListener[] listeners = (DataListener[])dataListeners.toArray(new DataListener[dataListeners.size()]);
      for(int i = 0; i < listeners.length; i++) {
         data.mark();
         ((DataListener)listeners[i]).frameData(data);
         data.reset();
      }
   }

   protected void fireTrackingDataReceived(ByteBuffer data) {
      ImageSystem.LOG.info("Received: Tracking Info (" + (data.get(data.position()) & 0xFF) + ")");
      DataListener[] listeners = (DataListener[])dataListeners.toArray(new DataListener[dataListeners.size()]);
      for(int i = 0; i < listeners.length; i++) {
         data.mark();
         ((DataListener)listeners[i]).trackingData(data);
         data.reset();
      }
   }

   private void sendRequest(byte[] data) throws IOException {
      sendRequest(data, 0, data.length);
   }

   private void sendRequest(byte[] data, int off, int len) throws IOException {
      out.write(data, off, len);
      out.write((byte)'\n');
      out.flush();

      StringBuffer builder = new StringBuffer("Sending: ");
      for(int i = off; i < len; i++) {
         builder.append((char)data[i]);
      }
      ImageSystem.LOG.info(builder.toString());
   }

   public void sendSetRegisters(Map registers) throws IOException {
      ByteBuffer buffer = ByteBuffer.allocate(60);
      buffer.put((byte)'C')
            .put((byte)'M');

      Set keys = registers.keySet();
      for(Iterator i = keys.iterator(); i.hasNext();) {
         String key = i.next().toString();
         buffer.put((byte)' ').put(key.getBytes())
               .put((byte)' ').put(registers.get(key).toString().getBytes());
      }

      buffer.put((byte)'\r');
      buffer.flip();
      sendRequest(buffer.array(), 0, buffer.remaining());
   }

   public void sendDisableTracking() throws IOException {
      sendRequest(new byte[] { (byte)'D', (byte)'T', (byte)'\r'} );
   }

   public void sendDumpFrame() throws IOException {
      sendRequest(new byte[] { (byte)'D', (byte)'F', (byte)'\r'} );
   }

   public void sendEnableTracking() throws IOException {
      sendRequest(new byte[] { (byte)'E', (byte)'T', (byte)'\r'} );
   }

   public void sendGetVersion() throws IOException {
      sendRequest(new byte[] { (byte)'G', (byte)'V', (byte)'\r'} );
   }

   public void sendPing() throws IOException {
      sendRequest(new byte[] { (byte)'P', (byte)'G', (byte)'\r'} );
   }

   public void sendReset() throws IOException {
      sendRequest(new byte[] { (byte)'R', (byte)'S', (byte)'\r'} );
   }

   public void sendColorMap(int[] map) throws IOException {
      ByteBuffer buffer = ByteBuffer.allocate(2 + 4 * map.length + 1);
      buffer.put("SM".getBytes());
      for(int i = 0; i < map.length; i++) {
         buffer.put((" " + map[i] + "").getBytes());
      }
      buffer.put("\r".getBytes());
      buffer.flip();
      sendRequest(buffer.array(), 0, buffer.remaining());
   }

   public void sendSetColorMap(int[] red, int[] green, int[] blue) throws IOException {
      ByteBuffer buffer = ByteBuffer.allocate(2 + 4 * red.length * 3 + 1);
      buffer.put("SM".getBytes());
      for(int i = 0; i < red.length; i++) {
         buffer.put((" " + red[i] + "").getBytes());
      }
      for(int i = 0; i < green.length; i++) {
         buffer.put((" " + green[i] + "").getBytes());
      }
      for(int i = 0; i < blue.length; i++) {
         buffer.put((" " + blue[i] + "").getBytes());
      }
      buffer.put("\r".getBytes());
      buffer.flip();
      sendRequest(buffer.array(), 0, buffer.remaining());
   }

   private static class SerialEventHandler implements SerialPortEventListener {

      private Device device;

      public SerialEventHandler(Device device) {
         this.device = device;
      }

      public void serialEvent(SerialPortEvent spe) {

         ImageSystem.LOG.finest("Serial Event: " + spe.toString());

         if(spe.getEventType() == SerialPortEvent.DATA_AVAILABLE) {

            try {

               do {

                  byte[] data = new byte[8192];

                  int bytesRead = device.getInputStream().read(data);
                  if(bytesRead > 0) {
                     // take only bytesRead - 1 to remove the
                     // terminating character '\r' or 0xFF
                     EventQueue.invokeLater(new GUITask(device, ByteBuffer.wrap(data, 0, bytesRead - 1)));
                  }
                  ImageSystem.LOG.finest("Bytes Read: " + bytesRead);

               } while(device.getInputStream().available() > 0);

            } catch(IOException ioe) {
               ImageSystem.LOG.severe(ioe.getMessage());
               device.disconnect();
            }

         }
      }

   }

   private static class GUITask implements Runnable {
      private ByteBuffer data;
      private Device device;

      public GUITask(Device device, ByteBuffer data) {
         this.device = device;
         this.data = data;
      }

      public void run() {
         device.handleData(data);
      }

   }

   /**
    * Only used for debugging purposes, this will simulate an ACK coming from the device.
    */
   public void simulateACK() {
      fireACKReceived();
   }

   /**
    * Only used for debugging purposes, this will simulate an NCK coming from the device.
    */
   public void simulateNCK() {
      fireNCKReceived();
   }

}
