package img.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.nio.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

import img.device.event.*;

import img.lang.*;

public class JCaptureInternalFrame extends JInternalFrame {

   private static final Format DATE_FORMAT;

   static {
      DATE_FORMAT = new SimpleDateFormat("MM/dd/yyyy hh:mm:ss");
   }

   private JMessagePanel messageP;
   private JColorMapPanel colorMapP;
   private JCapturePanel captureP;
   private DataListener dataHandler;

   private JLabel redValueL;
   private JLabel greenValueL;
   private JLabel blueValueL;

   private boolean fromCamera;

   public JCaptureInternalFrame(JMessagePanel messageP) {
      this(messageP, null);
   }

   public JCaptureInternalFrame(JMessagePanel messageP, JColorMapPanel colorMapP) {
      this(messageP, colorMapP, null);
   }

   public JCaptureInternalFrame(JMessagePanel messageP, JColorMapPanel colorMapP, File file) {
      super("Capture Frame" + ((file == null) ? ": " + DATE_FORMAT.format(new Date()) 
                                              : " " + file.toString()), 
            true, true, true, true);

      this.fromCamera = file == null;

      this.messageP = messageP;
      this.colorMapP = colorMapP;

      setDefaultCloseOperation(DISPOSE_ON_CLOSE);
      setJMenuBar(createMenuBar());

      redValueL = new JLabel("0");
      greenValueL = new JLabel("0");
      blueValueL = new JLabel("0");

      if(file != null) {
         try {
            captureP = new JCapturePanel(file);
         } catch(IOException ioe) {
            ImageSystem.LOG.severe(ioe.getMessage());
            ioe.printStackTrace();
         }
      } else {
         captureP = new JCapturePanel();
      }

      captureP.addMouseMotionListener(new MouseMotionHandler());

      Box southBox = new Box(BoxLayout.X_AXIS);
      southBox.setBorder(new EmptyBorder(5, 5, 5, 5));

      southBox.add(Box.createHorizontalGlue());
      southBox.add(new JLabel("Red: "));
      southBox.add(redValueL);
      southBox.add(Box.createHorizontalStrut(5));
      southBox.add(new JLabel("Green: "));
      southBox.add(greenValueL);
      southBox.add(Box.createHorizontalStrut(5));
      southBox.add(new JLabel("Blue: "));
      southBox.add(blueValueL);
      southBox.add(Box.createHorizontalGlue());

      getContentPane().add(captureP, BorderLayout.CENTER);
      getContentPane().add(southBox, BorderLayout.SOUTH);

      if(colorMapP != null) {
         captureP.addMouseListener(new MouseHandler());
      }

   }

   private JMenuBar createMenuBar() {
      JMenuBar menubar = new JMenuBar();

      JMenu fileM = new JMenu("File");

      try {
         fileM.add(new ProxyAction(this, "save", "Save", 's'));
         fileM.add(new ProxyAction(this, "pack", "Reset Size", 'r'));
         fileM.addSeparator();
         fileM.add(new ProxyAction(this, "dispose", "Exit", 'x'));
      } catch(Exception e) {
      }

      menubar.add(fileM);

      return menubar;
   }

   public void save() {

      int option = ImageSystem.FILE_CHOOSER.showSaveDialog(getDesktopPane().getRootPane());
      if(option == JFileChooser.APPROVE_OPTION) {
         try {
            File file = ImageSystem.FILE_CHOOSER.getSelectedFile();
            if(!file.getName().toLowerCase().endsWith(ImageSystem.BAYER_FILE_EXT)) {
               file = new File(file.getName() + "." + ImageSystem.BAYER_FILE_EXT);
            }
            captureP.saveBayer(file);
         } catch(IOException ioe) {
            ioe.printStackTrace();
            ImageSystem.LOG.severe(ioe.getMessage());
         }
      }
   }

   public void setVisible(boolean visible) {
      if(fromCamera) {
         if(visible) {
            try {
               dataHandler = new DumpFrameHandler();
               ImageSystem.DEVICE.addDataListener(dataHandler);
               getDesktopPane().getRootPane().getGlassPane().setVisible(true);
               ImageSystem.DEVICE.sendDumpFrame();
               messageP.append("Capture Snapshot");
            } catch(IOException ioe) {
               ImageSystem.DEVICE.removeDataListener(dataHandler);
               ImageSystem.LOG.severe(ioe.getMessage());
               messageP.append("Capture not sent");
            }
         } else {
            ImageSystem.DEVICE.removeDataListener(dataHandler);
         }
      }

      super.setVisible(visible);
   }

   private class DumpFrameHandler extends DataAdapter {

      private int frameCount;

      public DumpFrameHandler() {
      }

      public void ack() {
         frameCount = 0;
      }

      public void nck() {
         getDesktopPane().getRootPane().getGlassPane().setVisible(false);
         ImageSystem.DEVICE.removeDataListener(this);
         JOptionPane.showMessageDialog(messageP.getRootPane(), "Capture NCK Received", "NCK Received", JOptionPane.ERROR_MESSAGE);
      }

      public void frameData(ByteBuffer data) {

         frameCount++;
         captureP.setRow(data.get() & 0xFF, data, frameCount == 0x48);

         if(frameCount == 0x48) {
            ImageSystem.DEVICE.removeDataListener(this);
            getDesktopPane().getRootPane().getGlassPane().setVisible(false);
         }

      }

   }

   private class MouseHandler extends MouseAdapter {

      private JPopupMenu popupM;
      private int color;

      public void mouseReleased(MouseEvent me) {
         if(SwingUtilities.isRightMouseButton(me)) {
            Dimension size = captureP.getSize();
            Insets insets = captureP.getInsets();
            Dimension preferredSize = captureP.getPreferredSize();

            Point mouse = me.getPoint();

            double scaleX = size.width / (double)(5 + ImageSystem.IMAGE_WIDTH + 10 + ImageSystem.IMAGE_WIDTH + 5);
            double scaleY = size.height / (double)(5 + ImageSystem.IMAGE_HEIGHT + 5);
            double scale = Math.min(scaleX, scaleY);

            int imageX = (int)(insets.left + ((5 + ImageSystem.IMAGE_WIDTH + 10) * scale));
            int imageY = (int)(insets.top + (5 * scale));
            int imageWidth = (int)(ImageSystem.IMAGE_WIDTH * scale);
            int imageHeight = (int)(ImageSystem.IMAGE_HEIGHT * scale);

            if(((mouse.x >= imageX) && (mouse.x < (imageX + imageWidth))) && 
               ((mouse.y >= imageY) && (mouse.y < (imageY + imageHeight)))) {

               scale = 1 / scale;

               int x = (int)Math.round((mouse.x * scale) - (insets.left + 5 + ImageSystem.IMAGE_WIDTH + 10));
               int y = (int)Math.round((mouse.y * scale) - (insets.top + 5));

               color = captureP.getRGB(x, y);

               if(popupM == null) {
                  popupM = new JPopupMenu();
                  try {
                     /*
                     JMenu colorM = new JMenu("Color Map Column");
                     ActionListener handler = new ActionHandler();

                     int index = 1;
                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));

                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));
                     colorM.add(createButton(index++, handler));

                     popupM.add(colorM);
                     */

                     JMenuItem colorMapMI = new JMenuItem("Add to Color Map");
                     colorMapMI.addActionListener(new ActionHandler());
                     popupM.add(colorMapMI);
                  } catch(Exception e) {
                  }
               }

               popupM.show((Component)me.getSource(), me.getX(), me.getY());

            }

         }
      }

      private JMenuItem createButton(int index, ActionListener listener) {
         JMenuItem indexMI = new JMenuItem("" + index);
         indexMI.addActionListener(listener);
         return indexMI;
      }

      public class ActionHandler implements ActionListener {

         private JPanel displayP = null;
         private JRadioButton[] colRB = null;

         public void actionPerformed(ActionEvent ae) {

            int column = 0;
            int option = JOptionPane.YES_OPTION;

            do {

               option = JOptionPane.YES_OPTION;
               column = getColorMapColumn();

               if(!colorMapP.isColumnClear(column)) {


                  option = JOptionPane.showConfirmDialog(getRootPane(),
                                                            "Column " + column + " is already set. Overwrite current value?",
                                                            "Overwrite current column value?",
                                                            JOptionPane.YES_NO_CANCEL_OPTION,
                                                            JOptionPane.QUESTION_MESSAGE);
                  
                  
               }

            } while(option == JOptionPane.NO_OPTION);

            if(option == JOptionPane.YES_OPTION) {
               SwingUtilities.getWindowAncestor(colorMapP).setVisible(true);
               colorMapP.setColor(column, color);
            }

         }

         private int getColorMapColumn() {
            if(displayP == null) {
               displayP = new JPanel(new BorderLayout());

               JPanel selectColP = new JPanel();

               ButtonGroup bg = new ButtonGroup();

               colRB = new JRadioButton[8];
               for(int i = 0; i < colRB.length; i++) {
                  colRB[i] = new JRadioButton((i + 1) + "");
                  bg.add(colRB[i]);
                  selectColP.add(colRB[i]);
               }

               colRB[0].setSelected(true);

               displayP.add(new JLabel("Select Color Map Column:"), BorderLayout.NORTH);
               displayP.add(selectColP, BorderLayout.SOUTH);
            }

            int option = JOptionPane.showConfirmDialog(getRootPane(),
                                                       displayP,
                                                       "Add to Color Map Column",
                                                       JOptionPane.OK_CANCEL_OPTION,
                                                       JOptionPane.QUESTION_MESSAGE);

            int selected = -1;

            if(option == JOptionPane.OK_OPTION) {
               for(int i = 0; selected == -1 && i < colRB.length; i++) {
                  if(colRB[i].isSelected()) {
                     selected = i;
                  }
               }

            }

            return selected;

         }

      }

   }

   private class MouseMotionHandler implements MouseMotionListener {
      private Robot robot;

      public MouseMotionHandler() {
         try {
            robot = new Robot();
         } catch(AWTException awte) {
            ImageSystem.LOG.warning(awte.getMessage());
            awte.printStackTrace();
         }
      }

      public void mouseMoved(MouseEvent me) {
         Dimension size = captureP.getSize();
         Insets insets = captureP.getInsets();
         Dimension preferredSize = captureP.getPreferredSize();

         Point mouse = me.getPoint();

         double scaleX = size.width / (double)preferredSize.width;
         double scaleY = size.height / (double)preferredSize.height;
         double scale = Math.min(scaleX, scaleY);

         int imageX = (int)(insets.left + ((5 + ImageSystem.IMAGE_WIDTH + 10) * scale));
         int imageY = (int)(insets.top + (5 * scale));
         int imageWidth = (int)(ImageSystem.IMAGE_WIDTH * scale);
         int imageHeight = (int)(ImageSystem.IMAGE_HEIGHT * scale);

         if(((mouse.x >= imageX) && (mouse.x < (imageX + imageWidth))) && 
            ((mouse.y >= imageY) && (mouse.y < (imageY + imageHeight)))) {

            scale = 1 / scale;

            int x = (int)Math.round((mouse.x * scale) - (insets.left + 5 + ImageSystem.IMAGE_WIDTH + 10));
            int y = (int)Math.round((mouse.y * scale) - (insets.top + 5));

            int color = captureP.getRGB(x, y);

            redValueL.setText(((color & 0xFF0000) >> 16) + "");
            greenValueL.setText(((color & 0x00FF00) >> 8) + "");
            blueValueL.setText(((color & 0x0000FF) >> 0) + "");
         }
      }

      public void mouseDragged(MouseEvent me) {
      }
   }

}
