package log.util;

import java.util.*;
import java.util.logging.*;

/***********************************************************************
 * Defines Comparators to sort the log records.
 */
public class LogComparators implements Comparator {

   /**
    * Comparator used to sort the Log Records by the Log Level.
    */
   public static final Comparator LEVEL_COMPARATOR;

   /**
    * Comparator used to sort the Log Records by the Date received.
    */
   public static final Comparator DATE_COMPARATOR;

   /**
    * Comparator used to sort the Log Records by the contents of the Message.
    */
   public static final Comparator MESSAGE_COMPARATOR;

   /**
    * Constant used to create a Level Comparator
    */
   private static final int LEVEL_TYPE   = 0x00;

   /**
    * Constant used to create a Date Comparator
    */
   private static final int DATE_TYPE    = 0x01;

   /**
    * Constant used to create a Message Comparator
    */
   private static final int MESSAGE_TYPE = 0x02;

   static {

      LEVEL_COMPARATOR = new LogComparators(LEVEL_TYPE);
      DATE_COMPARATOR = new LogComparators(DATE_TYPE);
      MESSAGE_COMPARATOR = new LogComparators(MESSAGE_TYPE);

   }

   /**
    * The type of Comparator to use.
    */
   private int type;

   /**
    * Create a Log Comparator of the given type.
    * @param type The type of Comparator to use.
    */
   private LogComparators(int type) {
      this.type = type;
   }

   /**
    * Compare object to object1.
    * @param object The first object to be compared.
    * @param object1 The second object to be compared.
    * @return A value of -1 if object is less than object1, 0 if object
    * and object1 are equal, +1 if object is greater than object1.
    * @throws ClassCastException if either object or object1 is not a
    * LogRecord
    */
   public int compare(Object object, Object object1) {

      LogRecord rec1 = (LogRecord)object;
      LogRecord rec2 = (LogRecord)object1;

      int value = 0;

      switch(type) {
         // sort by the log's Level
         case LEVEL_TYPE:
            value = rec1.getLevel().intValue() - rec2.getLevel().intValue();

            // if the level's are the same, the sort by the sequence number
            if(value == 0) {
               value = compareLong(rec1.getSequenceNumber(),
                                   rec2.getSequenceNumber());
            }
            break;

         // sort by the log's Time
         case DATE_TYPE:
            value = compareLong(rec1.getMillis(), rec2.getMillis());

            // if the time is the same, the sort by the sequence number
            if(value == 0) {
               value = compareLong(rec1.getSequenceNumber(),
                                   rec2.getSequenceNumber());
            }
            break;

         // sort by the log's Message
         case MESSAGE_TYPE:
            value = rec1.getMessage().compareToIgnoreCase(rec2.getMessage());

            // if the messages are the same, the sort by the sequence number
            if(value == 0) {
               value = compareLong(rec1.getSequenceNumber(),
                                   rec2.getSequenceNumber());
            }
            break;
      }

      return value;

   }

   /**
    * Since the compare method returns an int value, by subtracting two
    * long values, the result is a long.  This function will subtract the
    * two long values and return an int.
    * @param value1 The first value to be compared.
    * @param value2 The second value to be compared
    * @return A value of -1 if value1 is less than value2, 0 if value1 and
    * value2 are equal, +1 if value1 is greater than value2.
    */
   private int compareLong(long value1, long value2) {
      long diff = value1 - value2;

      if(diff < 0) {
         return -1;
      } else if(diff > 0) {
         return +1;
      } else { // if(diff == 0) {
         return 0;
      }
   }

}
